/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.solr.handler.component;

import org.apache.solr.common.SolrException;
import org.apache.solr.common.params.CollapseParams;
import org.apache.solr.common.params.SolrParams;
import org.apache.solr.common.util.NamedList;
import org.apache.solr.core.SolrCore;
import org.apache.solr.request.SolrQueryRequest;
import org.apache.solr.request.SolrQueryResponse;
import org.apache.solr.search.DocListAndSet;
import org.apache.solr.search.SolrIndexSearcher;
import org.apache.solr.search.fieldcollapse.AdjacentDocumentCollapser;
import org.apache.solr.search.fieldcollapse.DocumentCollapseResult;
import org.apache.solr.search.fieldcollapse.DocumentCollapser;
import org.apache.solr.search.fieldcollapse.NonAdjacentDocumentCollapser;
import org.apache.solr.search.fieldcollapse.collector.*;
import org.apache.solr.util.plugin.SolrCoreAware;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Collapse component is responsible for do field collapsing with an implementation of {@link org.apache.solr.search.fieldcollapse.DocumentCollapser},
 * that does the most of the work. Collapsing is activated by specifying <code>collapse.field</code> in the request
 * as parameter.
 * If this parameter is not specified it falls back to the {@link QueryComponent#process(ResponseBuilder)} method.
 * <br/><br/>
 * If the parameter <code>collapse.facet</code> with value <code>after</code> is specified,
 * it replaces the collapsed docSet from <code>DocListAndSet</code> with the uncollapsed docset, so that facetation
 * counts are based on the uncollapsed search result (which is usually desired behaviour)
 *
 * @version $Id: CollapseComponent.java 602341 2007-12-08 07:27:49Z ryan $
 * @since solr 1.3
 */
public class CollapseComponent extends QueryComponent implements SolrCoreAware {

  private List<CollapseCollectorFactory> collapseCollectorFactories = new ArrayList<CollapseCollectorFactory>();
  private List<String> collapseCollectorFactoryNames;

  /**
   * {@inheritDoc}
   */
  @Override
  @SuppressWarnings("unchecked")
  public void init(NamedList args) {
    super.init(args);
    collapseCollectorFactoryNames = (List<String>) args.get("collapseCollectorFactories");
  }

  /**
   * {@inheritDoc}
   */
  public void inform(SolrCore core) {
    if (collapseCollectorFactoryNames == null) {
      collapseCollectorFactories.add(new DocumentGroupCountCollapseCollectorFactory());
      collapseCollectorFactories.add(new FieldValueCountCollapseCollectorFactory());
      collapseCollectorFactories.add(new DocumentFieldsCollapseCollectorFactory());
      // ok is not have its done normally, but works for now... need something like loadSearchComponents() and addIfNotPresent()
      // but requires more changes outside field collapse code. 
      AggregateCollapseCollectorFactory aggregateCollapseCollectorFactory = new AggregateCollapseCollectorFactory();
      aggregateCollapseCollectorFactory.init(new NamedList());
      collapseCollectorFactories.add(aggregateCollapseCollectorFactory);
      return;
    }

    Map<String, CollapseCollectorFactory> registry = new HashMap<String, CollapseCollectorFactory>();
    core.initPlugins(registry, CollapseCollectorFactory.class);
    for (String name : collapseCollectorFactoryNames) {
      CollapseCollectorFactory factory = registry.get(name);
      if (factory == null) {
        throw new SolrException(SolrException.ErrorCode.BAD_REQUEST, "Unknown collapse collector factory: " + name);
      }

      collapseCollectorFactories.add(factory);
    }
  }

  /**
   * Parses some parameters from the request in preparation for the collapsing when collapsing is enabled otherwise
   * behaves like {@link super#prepare(ResponseBuilder)}.
   *
   * @param rb The response builder of the current search request
   * @throws IOException When index search problems occur
   */
  @Override
  public void prepare(ResponseBuilder rb) throws IOException {
    super.prepare(rb);
    rb.req.getContext().put("collapseRequest", resolveCollapseRequest(rb));
  }

  /**
   * Actually run the query
   */
  @Override
  public void process(ResponseBuilder rb) throws IOException {
    CollapseRequest collapseRequest = (CollapseRequest) rb.req.getContext().remove("collapseRequest");
    if (collapseRequest == null) {
      super.process(rb);
      return;
    }
    doProcess(rb, collapseRequest);
  }

  @Override
  public void handleResponses(ResponseBuilder rb, ShardRequest sreq) {
    super.handleResponses(rb, sreq);
    if ((sreq.purpose & ShardRequest.PURPOSE_GET_TOP_IDS) != 0) {
      mergeCollapseCounts(rb, sreq);
    }
  }

  // ================================================= Helpers =========================================================

  /**
   * Resolves the field collapse parameters from the request and constructs a <code>CollapseRequest</code> from it.
   *
   * @param rb The response builder that contains the request
   * @return a <code>CollapseRequest</code> based on the parameters in the request
   * @throws IOException When IO related problems occur
   */
  protected CollapseRequest resolveCollapseRequest(ResponseBuilder rb) throws IOException {
    SolrParams params = rb.req.getParams();
    if (params.get(CollapseParams.COLLAPSE_FIELD) == null) {
      return null;
    }

    String facet = params.get(CollapseParams.COLLAPSE_FACET);
    CollapseParams.CollapseFacet collapseFacet = (facet != null) ? CollapseParams.CollapseFacet.get(facet) : CollapseParams.CollapseFacet.AFTER;

    String type = params.get(CollapseParams.COLLAPSE_TYPE);
    CollapseParams.CollapseType collapseType = (type != null) ? CollapseParams.CollapseType.get(type) : CollapseParams.CollapseType.NORMAL;
    DocumentCollapser documentCollapser = constructCollapser(rb, collapseType);
    return new CollapseRequest(documentCollapser, collapseFacet);
  }

  /**
   * Executes a search and collapses the search results.
   *
   * @param rb The response builder of the current search request
   * @param collapseRequest The current collapse request
   * @throws IOException When index search problems occur
   */
  protected void doProcess(ResponseBuilder rb, CollapseComponent.CollapseRequest collapseRequest) throws IOException {
    SolrQueryResponse rsp = rb.rsp;
    SolrQueryRequest req = rb.req;
    SolrIndexSearcher searcher = req.getSearcher();
    DocumentCollapseResult collapseResult =
            collapseRequest.getCollapser().collapse(rb.getQuery(), rb.getFilters(), rb.getSortSpec().getSort());

    DocListAndSet results = searcher.getDocListAndSet(rb.getQuery(),
      collapseResult == null ? rb.getFilters() : null,
      collapseResult.getCollapsedDocset(),
      rb.getSortSpec().getSort(),
      rb.getSortSpec().getOffset(),
      rb.getSortSpec().getCount(),
      rb.getFieldFlags());

    //for getting the facet count BEFORE the collapsing, we must
    //get the doc. collection without filtering by the collapseResult.
    boolean facetAfterCollapse = (collapseRequest.getCollapseFacet() == CollapseParams.CollapseFacet.AFTER);
    if (!facetAfterCollapse) {
      results.docSet = collapseResult.getUnCollapsedDocset();
    }

    rb.setResults(results);
    rsp.add("collapse_counts", collapseRequest.getCollapser().getCollapseInfo(searcher, results.docList));
    rsp.add("response", results.docList);

    doFSV(rb, searcher);
    doPrefetch(rb);
  }

  /**
   * Returns the field collapser implementation based on the specified collapseType
   *
   * @param rb           The response builder of the current search request
   * @param collapseType The type of collapsing for the current search
   * @return collapser implementation
   * @throws IOException When index search problems occur
   */
  protected DocumentCollapser constructCollapser(ResponseBuilder rb, CollapseParams.CollapseType collapseType) throws IOException {
    switch (collapseType) {
      case ADJACENT:
        return new AdjacentDocumentCollapser(rb, collapseCollectorFactories);
      case NORMAL:
        return new NonAdjacentDocumentCollapser(rb, collapseCollectorFactories);
      default:
        throw new UnsupportedOperationException("Method not implemented");
    }
  }

  /**
   * Merges the collapse responses from the shards into one distributed collapse response.
   *
   * @param rb The response builder
   * @param sreq The shard request
   */
  protected void mergeCollapseCounts(ResponseBuilder rb, ShardRequest sreq) {
    String collapseField = "";
		NamedList collapseResults = new NamedList<Object>();

		for (ShardResponse srsp : sreq.responses) {
			NamedList collapseInfo = (NamedList<Object>) srsp.getSolrResponse().getResponse().get("collapse_counts");
			if (collapseInfo != null) {
				collapseField = (String) collapseInfo.get("field");
				NamedList collapseResult = (NamedList) collapseInfo.get("results");
        for (int i = 0; i < collapseResult.size(); i++) {
          String collapseGroupId = collapseResult.getName(i);
          ShardDoc sdoc = rb.resultIds.get(collapseGroupId);
				  if (sdoc != null) {
            collapseResults.add(collapseGroupId, collapseResult.getVal(i));
          }
        }
			}
		}

		if (collapseResults.size() > 0) {
			NamedList<Object> result = new NamedList<Object>();
			result.add("field", collapseField);
			result.add("results", collapseResults);
			rb.rsp.add("collapse_counts", result);
		}
  }

  /////////////////////////////////////////////
  ///  SolrInfoMBean
  ////////////////////////////////////////////

  @Override
  public String getDescription() {
    return "Field Collapsing";
  }

  @Override
  public String getVersion() {
    return "";
  }

  @Override
  public String getSourceId() {
    return "";
  }

  @Override
  public String getSource() {
    return "";
  }

  @Override
  public URL[] getDocs() {
    return null;
  }


  // ============================================ Inner Classes ========================================================

  static class CollapseRequest {

    private final DocumentCollapser collapser;
    private final CollapseParams.CollapseFacet collapseFacet;

    private CollapseRequest(DocumentCollapser collapser, CollapseParams.CollapseFacet collapseFacet) {
      this.collapser = collapser;
      this.collapseFacet = collapseFacet;
    }

    public DocumentCollapser getCollapser() {
      return collapser;
    }

    public CollapseParams.CollapseFacet getCollapseFacet() {
      return collapseFacet;
    }

  }
}
