/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.solr.search.fieldcollapse;

import org.apache.lucene.search.FieldCache;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.Sort;
import org.apache.lucene.util.OpenBitSet;
import org.apache.solr.common.params.CollapseParams;
import org.apache.solr.common.params.SolrParams;
import org.apache.solr.common.util.NamedList;
import org.apache.solr.handler.component.ResponseBuilder;
import org.apache.solr.request.SolrQueryRequest;
import org.apache.solr.schema.IndexSchema;
import org.apache.solr.schema.SchemaField;
import org.apache.solr.search.*;
import org.apache.solr.search.fieldcollapse.collector.CollapseCollector;
import org.apache.solr.search.fieldcollapse.collector.CollapseCollectorFactory;
import org.apache.solr.search.fieldcollapse.collector.CollapseContext;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Base <code>Collapser</code> implementation
 */
public abstract class AbstractDocumentCollapser implements DocumentCollapser {

  /**
   * Field to use to collapse results. Parameter.
   */
  protected final String collapseField;

  /**
   * Number of documents with the same value for collapseField after which
   * collapsing kicks in. Parameter.
   */
  protected final int collapseThreshold;

  /** Collapse State *********************************************** */

  /**
   * Maximum number of documents to process during field collapsing.
   * Parameter.
   */
  protected final int collapseMaxDocs;

  /**
   * Whether to include debug information in the request
   */
  protected boolean includeDebugInformation;

  /**
   * Maxmimum size for a HashDocSet.
   */
  protected final int hashMaxSize;

  /**
   * Buffer for collecting documents. Gets turned into different types of
   * DocSet depending on the number of documents we end up with.
   */
  protected int[] docbuf;

  /**
   * Number of documents in docbuf.
   */
  protected int docbufSize = 0;

  /**
   * Maximum document id currently in docbuf. Only valid while docbufSize <
   * hashMaxSize.
   */
  protected int docbufMaxDoc = 0;

  /**
   * Bitset representation of docbuf. Gets created when docbufSize >=
   * hashMaxSize.
   */
  protected OpenBitSet docbufBitSet;

  /**
   * The result of executing the query and filter queries, without collapsing.
   */
  protected DocSet uncollapsedDocSet;

  protected FieldCache.StringIndex fieldValues;
  protected final SolrIndexSearcher searcher;
  protected final List<CollapseCollector> collectors;


  protected final int flags;

  protected final CollapseContext collapseContext;

  private final List<CollapseCollectorFactory> collapseCollectorFactories;

  /**
   * Debug Information ********************************************
   */
  protected long timeCollapsing = 0;
  protected long timeConvertToBitSet = 0;
  protected long timeCreateDocSet = 0;
  protected long timeCreateCollapseInfo = 0;
  protected long timeCreateUncollapedDocset = 0;
  protected String debugDocSetInfo = "unknown";


  // ============================================ Constructors =======================================================

  /**
   * Constructs a base collapser.
   *
   * @param rb The response builder
   * @param collapseCollectorFactories A list of collapse collector factories configured for the document collapser
   * @throws IOException if index searcher related problems occur
   */
  protected AbstractDocumentCollapser(ResponseBuilder rb, List<CollapseCollectorFactory> collapseCollectorFactories) throws IOException {
    this.collapseCollectorFactories = collapseCollectorFactories;

    // Allocate data structures
    hashMaxSize = rb.req.getCore().getSolrConfig().hashDocSetMaxSize;
    docbuf = new int[hashMaxSize];
    this.searcher = rb.req.getSearcher();
    this.flags = rb.getFieldFlags();

    // parsing Solr parameters
    SolrParams params = rb.req.getParams();
    collapseField = params.required().get(CollapseParams.COLLAPSE_FIELD);
    Integer ct = params.getInt(CollapseParams.COLLAPSE_THRESHOLD);
    if (ct == null) {
      ct = params.getInt(CollapseParams.COLLAPSE_MAX);
    }
    collapseThreshold = ct != null ? ct : 1;
    checkCollapseField(rb.req.getSchema());

    int collapseMaxDocs = params.getInt(CollapseParams.COLLAPSE_MAXDOCS, 0);
    this.collapseMaxDocs = collapseMaxDocs <= 0 ? searcher.maxDoc() : collapseMaxDocs;
    includeDebugInformation = params.getBool(CollapseParams.COLLAPSE_DEBUG ,false);

    collapseContext = new CollapseContext();
    collectors = initializeCollapseCollectors(rb.req);
  }


  // ======================================= Public API methods ========================================================

  /**
   * {@inheritDoc}
   */
  public DocumentCollapseResult collapse(Query mainQuery, List<Query> filterQueries, Sort sort) throws IOException {
    long startTime = System.currentTimeMillis();
    doQuery(mainQuery, filterQueries, sort);
    timeCreateUncollapedDocset = System.currentTimeMillis() - startTime;
    fieldValues = FieldCache.DEFAULT.getStringIndex(searcher.getReader(), collapseField);
    doCollapsing(uncollapsedDocSet, fieldValues);
    return createDocumentCollapseResult();
  }

  /**
   * {@inheritDoc}
   */
  public NamedList<Object> getCollapseInfo(SolrIndexSearcher searcher, DocList docs) throws IOException {
    if (uncollapsedDocSet == null) {
      throw new IllegalStateException("Invoke the collapse method before invoking getCollapseInfo method");
    }

    long startTime = System.currentTimeMillis();

    NamedList<Object> collapseResponse = new NamedList<Object>();
    collapseResponse.add("field", collapseField);

    NamedList<Object> collapseResult = new NamedList<Object>();
    for (CollapseCollector collapseCollector : collectors) {
      collapseCollector.getResult(collapseResult, docs, collapseContext);
    }
    collapseResponse.add("results", collapseResult);

    if (includeDebugInformation) {
      collapseResponse.add("debug", getDebugInfo());
    }

    timeCreateCollapseInfo = System.currentTimeMillis() - startTime;
    return collapseResponse;
  }


  // ================================================= Helpers =======================================================

  /**
   * Executes implementation specific collapsing algorithm.
   *
   * @param uncollapsedDocset The uncollapsed DocSet
   * @param values The field values
   */
  protected abstract void doCollapsing(DocSet uncollapsedDocset, FieldCache.StringIndex values);

  /**
   * @return timing information for field collapsing process.
   */
  protected NamedList getDebugInfo() {
    long totalTime = timeCreateUncollapedDocset + timeCollapsing + timeCreateCollapseInfo + timeConvertToBitSet + timeCreateDocSet;
    NamedList<Object> namedList = new NamedList<Object>();
    namedList.add("Docset type", debugDocSetInfo);
    namedList.add("Total collapsing time(ms)", totalTime);
    namedList.add("Create uncollapsed docset(ms)", timeCreateUncollapedDocset);
    namedList.add(String.format("%s collapsing time(ms)", getClass().getSimpleName()), timeCollapsing);
    namedList.add("Creating collapseinfo time(ms)", timeCreateCollapseInfo);
    namedList.add("Convert to bitset time(ms)", timeConvertToBitSet);
    namedList.add("Create collapsed docset time(ms)", timeCreateDocSet);
    return namedList;
  }

  /**
   * Creates the document collapse result based on the uncollapsed docset and the collapsed docset.
   *
   * @return he document collapse result
   */
  protected DocumentCollapseResult createDocumentCollapseResult() {
    long startTime = System.currentTimeMillis();
    DocSet result = (docbufBitSet != null) ? new BitDocSet(docbufBitSet) : new HashDocSet(docbuf, 0, docbufSize);
    timeCreateDocSet = System.currentTimeMillis() - startTime;
    debugDocSetInfo = result.getClass().getSimpleName() + "(" + docbufSize + ")";
    return new DocumentCollapseResult(result, uncollapsedDocSet);
  }

  /**
   * Adds a document to the internal document buffer.
   *
   * @param doc The lucene identifier of the document to add
   */
  protected void addDoc(int doc) {
    // If we have less than hashMaxSize documents, just
    // keep adding them to docbuf. We will turn them into
    // a HashDocSet later.
    if (docbufSize < hashMaxSize) {
      docbuf[docbufSize] = doc;
      if (doc > docbufMaxDoc) {
        docbufMaxDoc = doc;
      }
    } else {
      // We have exceeded hashMaxSize. Allocate a bit set
      // if we don't have one yet, then add to that.
      if (docbufBitSet == null) {
        long startTime = System.currentTimeMillis();
        docbufBitSet = new OpenBitSet(docbufMaxDoc + 1);
        for (int i = 0; i < docbufSize; i++) {
          docbufBitSet.fastSet(docbuf[i]);
        }
        timeConvertToBitSet = System.currentTimeMillis() - startTime;
      }
      docbufBitSet.set(doc);
    }
    docbufSize++;
  }

  protected void doQuery(Query mainQuery, List<Query> filterQueries, Sort sort) throws IOException {
    uncollapsedDocSet = searcher.getDocList(mainQuery, filterQueries, sort, 0, collapseMaxDocs, flags);
  }

  /**
   * Method that checks if proper field collapsing is actually possible with the current collapse field.
   * If the collapse field does not meet the proper requirements a runtime exception is thrown.
   * A runtime exception is thrown under the following circumstances:
   * <ul>
   *  <li> When the collapse field does not exists in the schema
   *  <li> When the collapse field is multivalued in the schema
   *  <li> When the collapse field is tokenized in the schema
   * </ul>
   *
   * For example when a field is tokenized, only the last token of the field can be retrieved from the fieldcache. This
   * results in field collapsing only on the last token of a field value instead of the complete field value.
   *
   * When the field values from a multivalued field are returned from the <code>FieldCache</code> then an exception may
   * be thrown. This happens when there are more terms in a field than documents.
   *
   * @param schema The index schema
   */
  protected void checkCollapseField(IndexSchema schema) {
    SchemaField schemaField = schema.getFieldOrNull(collapseField);
    if (schemaField == null) {
      throw new RuntimeException("Could not collapse, because collapse field does not exist in the schema.");
    }

    if (schemaField.multiValued()) {
      throw new RuntimeException("Could not collapse, because collapse field is multivalued");
    }

    if (schemaField.getType().isTokenized()) {
      throw new RuntimeException("Could not collapse, because collapse field is tokenized");
    }
  }

  protected List<CollapseCollector> initializeCollapseCollectors(SolrQueryRequest request) {
    List<CollapseCollector> collectors = new ArrayList<CollapseCollector>();
    for (CollapseCollectorFactory factory : collapseCollectorFactories) {
      CollapseCollector collapseCollector = factory.createCollapseCollector(request);
      if (collapseCollector != null) {
        if (collectors.isEmpty()) {
          collectors.add(new DocumentHeadCollapseGroupCollector());
        }

        collectors.add(collapseCollector);
      }
    }
    return collectors;
  }

  // =========================================== Getters / Setters ===================================================

 /**
   * @return field to use to collapse results
   */
  public String getCollapseField() {
    return collapseField;
  }

  /**
   * @return maximum number of documents to process during field collapsing
   */
  public int getCollapseMaxDocs() {
    return collapseMaxDocs;
  }

  /**
   * @return number of documents with the same value for collapseField after which collapsing kicks in.
   */
  public int getCollapseThreshold() {
    return collapseThreshold;
  }


  //============================================== Inner classes =======================================================

  // Collector that associates a collapse group with the document head identifier.
  // This is information that many collapse collectors utilize.
  // This collector has to be the first collector. Order does matter.
  public static class DocumentHeadCollapseGroupCollector implements CollapseCollector {

    public void documentCollapsed(int docId, CollapseGroup collapseGroup, CollapseContext collapseContext) {
    }

    public void documentHead(int docHeadId, CollapseGroup collapseGroup, CollapseContext collapseContext) {
      collapseContext.getDocumentHeadCollapseGroupAssociation().put(docHeadId, collapseGroup);
    }

    public void getResult(NamedList result, DocList docs, CollapseContext collapseContext) {
    }
  }
}
