/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.solr.search.fieldcollapse;

import org.apache.lucene.search.FieldCache;
import org.apache.solr.handler.component.ResponseBuilder;
import org.apache.solr.search.DocIterator;
import org.apache.solr.search.DocSet;
import org.apache.solr.search.fieldcollapse.collector.CollapseCollector;
import org.apache.solr.search.fieldcollapse.collector.CollapseCollectorFactory;

import java.io.IOException;
import java.util.List;

/**
 * Adjacent collapsing behaviour. Has the same behaviour as {@link NonAdjacentDocumentCollapser} but with one extra constraint.
 * The collapsing happens in an adjacent manner. Meaning that only documents that are next to each other in a continuous
 * row and that have the same field value might be collapsed.
 */
public class AdjacentDocumentCollapser extends AbstractDocumentCollapser {

  /**
   * See {@link super#AbstractDocumentCollapser(org.apache.solr.handler.component.ResponseBuilder, List)}
   */
  public AdjacentDocumentCollapser(ResponseBuilder rb, List<CollapseCollectorFactory> collapseCollectorFactories) throws IOException {
    super(rb, collapseCollectorFactories);
  }


  // ================================================= Helpers =========================================================

  /**
   * Applies adjacent collapsing on the specified uncollapsedDocSet.
   *
   * @param uncollapsedDocset The uncollapsed docset
   * @param values            The fieldvalues the collapse on
   */
  protected void doCollapsing(DocSet uncollapsedDocset, FieldCache.StringIndex values) {
    int docCount = 0; // how many documents we have processed
    int repeatCount = 0; // how many times we have seen the same value in a
    int collapseCount = 0; // how many documents we have collapsed in this
    int collapseId = -1; // the document we're collapsing into
    String collapseValue = null;


    long startTime = System.currentTimeMillis();
    for (DocIterator i = uncollapsedDocset.iterator(); i.hasNext();) {
      int currentId = i.nextDoc();
      String currentValue = values.lookup[values.order[currentId]];
      
      // Initializing
      if (collapseValue == null) {
        repeatCount = 0;
        collapseCount = 0;
        collapseId = currentId;
        collapseValue = currentValue;

        // Collapse the document if the field value is the same and
        // we have a run of at least collapseThreshold uncollapsedDocset.
      } else if (collapseValue.equals(currentValue)) {
        if (++repeatCount >= collapseThreshold) {
          collapseCount++;
          for (CollapseCollector collector : collectors) {
            CollapseGroup valueToCollapse = new AdjacentCollapseGroup(collapseId, currentValue);
            collector.documentCollapsed(currentId, valueToCollapse, collapseContext);
          }
        } else {
          addDoc(currentId);
        }
      } else {
        addDoc(collapseId);
        if (collapseCount > 0) {
          CollapseGroup valueToCollapse = new AdjacentCollapseGroup(collapseId, collapseValue);
          for (CollapseCollector collector : collectors) {
            collector.documentHead(collapseId, valueToCollapse, collapseContext);
          }
        }

        repeatCount = 0;
        collapseCount = 0;
        collapseId = currentId;
        collapseValue = currentValue;
      }

      // Stop after collapseMaxDocs documents
      if (++docCount >= collapseMaxDocs) {
        break;
      }
    }

    if (collapseId != -1) {
      addDoc(collapseId);
    }

    if (collapseCount > 0) {
      CollapseGroup valueToCollapse = new AdjacentCollapseGroup(collapseId, collapseValue);
      for (CollapseCollector collector : collectors) {
          collector.documentHead(collapseId, valueToCollapse, collapseContext);
      }
    }

    timeCollapsing = System.currentTimeMillis() - startTime;
  }


  // ============================================ Inner Classes ========================================================

  private static class AdjacentCollapseGroup implements CollapseGroup {

    private final int firstDocumentId;
    private final String fieldValue;

    private AdjacentCollapseGroup(int firstDocumentId, String fieldValue) {
      this.firstDocumentId = firstDocumentId;
      this.fieldValue = fieldValue;
    }

    public String getKey() {
      return fieldValue;
    }

    @Override
    public boolean equals(Object o) {
      if (this == o) return true;
      if (o == null || getClass() != o.getClass()) return false;

      AdjacentCollapseGroup that = (AdjacentCollapseGroup) o;

      if (firstDocumentId != that.firstDocumentId) return false;
      if (fieldValue != null ? !fieldValue.equals(that.fieldValue) : that.fieldValue != null) return false;

      return true;
    }

    @Override
    public int hashCode() {
      int result = firstDocumentId;
      result = 31 * result + (fieldValue != null ? fieldValue.hashCode() : 0);
      return result;
    }
  }

}
