/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.solr.search.fieldcollapse;

import org.apache.lucene.search.Query;
import org.apache.lucene.search.Sort;
import org.apache.solr.common.util.NamedList;
import org.apache.solr.search.DocList;
import org.apache.solr.search.SolrIndexSearcher;

import java.io.IOException;
import java.util.List;

/**
 * A document collapser is responsible for removing documents that have the same value on a predefined field,
 * only the most relevant (based sorting and / or scoring) documents will <b>not</b> be removed.<br />
 * <br />
 * Besides that a document collapser also creates statistics about the collaped docset e.g. how many documents
 * were collapsed under the most relevant document and how many documents were collaped under a certain field value.
 */
public interface DocumentCollapser {

  /**
   * Executes a search based on the specified parameters
   * Returns a docset only containing the non collapsed documents also known as the document heads.
   *
   * @param mainQuery     The main query
   * @param filterQueries The filter queries
   * @param sort          The sort
   * @return a docset only containing the non collapsed documents
   * @throws IOException if index searcher related problems occur
   */
  DocumentCollapseResult collapse(Query mainQuery, List<Query> filterQueries, Sort sort) throws IOException;

  /**
   * Returns the statistics about the collapsed <code>DocSet</code>.
   * The following information is returned:
   * <ul>
   *  <li> The fieldname used during collapsing
   *  <li> A list of head document ids and how many documents were collapsed under it
   *  <li> A list of field values and how many documents were collapsed under under it. The field values are from the
   *       collapse field and
   *  <li> Debug information
   * </ul>
   *
   * This method may only be invoked when the {@link this#collapse(Query, List, Sort)} has been invoked.
   *
   * @param searcher The solr index searcher
   * @param docs     The doclist containing the results to be displayed
   * @return collapse counts for all documents in the specified docList.
   * @throws IOException if index searcher related problems occur
   */
  NamedList<Object> getCollapseInfo(SolrIndexSearcher searcher, DocList docs) throws IOException;

}
