/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.solr.search.fieldcollapse;

import org.apache.lucene.index.IndexReader;
import org.apache.lucene.search.*;
import org.apache.lucene.util.PriorityQueue;
import org.apache.solr.handler.component.ResponseBuilder;
import org.apache.solr.search.DocIterator;
import org.apache.solr.search.DocSet;
import org.apache.solr.search.fieldcollapse.collector.CollapseCollector;
import org.apache.solr.search.fieldcollapse.collector.CollapseCollectorFactory;
import org.apache.solr.util.DocSetScoreCollector;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Normal collapsing behaviour. Collapses all documents that have the same field value on the predefined field
 * and that have occured equal or more then collapse threshold.
 */
public class NonAdjacentDocumentCollapser extends AbstractDocumentCollapser {

  DocumentComparator documentComparator;

  /**
   * See {@link super#AbstractDocumentCollapser(org.apache.solr.handler.component.ResponseBuilder, List)}
   */
  public NonAdjacentDocumentCollapser(ResponseBuilder rb, List<CollapseCollectorFactory> collapseCollectorFactories) throws IOException {
    super(rb, collapseCollectorFactories);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  protected void doQuery(Query mainQuery, List<Query> filterQueries, Sort sort) throws IOException {
    final int maxDoc = searcher.maxDoc();
    if (containsSortOnScore(sort)) {
      DocSetScoreCollector docSetCollector = new DocSetScoreCollector(searcher.maxDoc());
      long startTime = System.currentTimeMillis();
      DocSet filter = searcher.getDocSet(filterQueries);
      uncollapsedDocSet = searcher.getDocSet(mainQuery, filter, docSetCollector);
      timeCreateUncollapedDocset = System.currentTimeMillis() - startTime;
      if (sort == null) {
        sort = new Sort(new SortField("score", SortField.SCORE, true));
      }

      float[] scores;
      if (uncollapsedDocSet instanceof DocSetScoreCollector.DelegateDocSet) {
        scores = ((DocSetScoreCollector.DelegateDocSet) uncollapsedDocSet).getScores();
      } else {
        scores = docSetCollector.getScores();
      }
      documentComparator = new DocumentComparator(sort, maxDoc, searcher.getIndexReader(), scores);
    } else {
      long startTime = System.currentTimeMillis();
      DocSet filter = searcher.getDocSet(filterQueries);
      uncollapsedDocSet = searcher.getDocSet(mainQuery, filter);
      timeCreateUncollapedDocset = System.currentTimeMillis() - startTime;

      documentComparator = new DocumentComparator(sort, maxDoc, searcher.getIndexReader());
    }
  }

  /**
   * {@inheritDoc}
   */
  protected void doCollapsing(DocSet uncollapsedDocset, FieldCache.StringIndex values) {
    long startTime = System.currentTimeMillis();
    int docCount = 0;
    // Keep how many documents we have processed the track of how many docs
    // with the same collapse value we have processed so far.
    Map<String, CollapsedDocumentGroup> collapsedDocs = new HashMap<String, CollapsedDocumentGroup>();

    for (DocIterator i  = uncollapsedDocset.iterator(); i.hasNext();) {
      int currentId = i.nextDoc();
      String currentValue = values.lookup[values.order[currentId]];

      // Get the last doc. and the total amount of docs. we have seen so
      // far for this collapsing value
      CollapsedDocumentGroup collapseDoc = collapsedDocs.get(currentValue);
      if (collapseDoc == null) {
        // new collapsing value => create a new record for it
        collapseDoc = new CollapsedDocumentGroup(0, 0, documentComparator, collapseThreshold);
        collapsedDocs.put(currentValue, collapseDoc);
      }
      // dropoutId has a value smaller than the smallest value in the queue and therefore it was removed from the queue
      Integer dropOutId = (Integer) collapseDoc.priorityQueue.insertWithOverflow(currentId);

      // check if we have reached the collapse threshold, if so start counting collapsed documents
      if (++collapseDoc.totalCount > collapseThreshold) {
        collapseDoc.collapsedDocuments++;
        if (dropOutId != null) {
          NonAdjacentCollapseGroup collapseValue = new NonAdjacentCollapseGroup(currentValue);
          for (CollapseCollector collector : collectors) {
            collector.documentCollapsed(dropOutId, collapseValue, collapseContext);
          }
        }
      }

      // Stop after collapseMaxDocs documents
      if (++docCount >= collapseMaxDocs) {
        break;
      }
    }

    // adding the head documents to the internal document buffer and
    // adding the collapsed counts per document head to the map
    for (String fieldValue : collapsedDocs.keySet()) {
      CollapsedDocumentGroup collapseDoc = collapsedDocs.get(fieldValue);
      if (collapseDoc.collapsedDocuments > 0) {
        NonAdjacentCollapseGroup collapseValue = new NonAdjacentCollapseGroup(fieldValue);
        for (CollapseCollector collector : collectors) {
          collector.documentHead((Integer) collapseDoc.priorityQueue.top(), collapseValue, collapseContext);
        }
      }
      Integer doc;
      while ((doc = (Integer) collapseDoc.priorityQueue.pop()) != null) {
        addDoc(doc);
      }
    }

    timeCollapsing = System.currentTimeMillis() - startTime;
  }


  // ================================================= Helpers =========================================================

  /**
   * Returns <code>true</code> if the sort contains a sortfield that sorts on score, otherwise <code>false</code>.
   *
   * @param sort The sort
   * @return <code>true</code> if the sort contains a sortfield that sorts on score, otherwise <code>false</code>
   */
  protected boolean containsSortOnScore(Sort sort) {
    if (sort == null) {
      return true; // means default sorting, which is sorting on score desc
    }

    for (SortField field : sort.getSort()) {
      if (field.getType() == SortField.SCORE) {
        return true;
      }
    }
    return false;
  }


  // ============================================ Inner Classes ========================================================

  private static class NonAdjacentCollapseGroup implements CollapseGroup {

    private final String fieldValue;

    private NonAdjacentCollapseGroup(String fieldValue) {
      this.fieldValue = fieldValue;
    }

    public String getKey() {
      return fieldValue;
    }

    @Override
    public boolean equals(Object o) {
      if (this == o) return true;
      if (o == null || getClass() != o.getClass()) return false;

      NonAdjacentCollapseGroup that = (NonAdjacentCollapseGroup) o;

      if (fieldValue != null ? !fieldValue.equals(that.fieldValue) : that.fieldValue != null) return false;

      return true;
    }

    @Override
    public int hashCode() {
      return fieldValue != null ? fieldValue.hashCode() : 0;
    }
  }

  /**
   * A <code>PriorityQueue</code> that maintaince order with a <code>DocumentComparator</code>.
   */
  public static class DocumentPriorityQueue extends PriorityQueue {

    private final DocumentComparator comparator;

    /**
     * Constructs a <code>DocumentPriorityQueue</code>
     *
     * @param comparator The <code>DocumentComparator</code> used for maintaining order in the queue
     * @param max        The maximum number of document identifiers in the queue (is equal to collapse thresold paramter)
     */
    public DocumentPriorityQueue(DocumentComparator comparator, int max) {
      this.comparator = comparator;
      initialize(max);
    }

    /**
     * {@inheritDoc}
     */
    protected boolean lessThan(Object a, Object b) {
      return comparator.compare((Integer) a, (Integer) b) < 0;
    }
  }

  /**
   * Compares two documents with each other.
   */
  public static class DocumentComparator {

    private final FieldComparator[] fieldComparators;
    private final boolean[] descending;
    private Scorer scorer;

    /**
     * Constructs a <code>DocumentComparator</code> by initializing the
     * {@link org.apache.lucene.search.FieldComparator}s and determining the sort orders.
     *
     * @param sort         The sort used for the creation of the FieldComparators.
     * @param numberOfHits The number of results in the pre-field-collapsed resultset
     * @param reader       The index reader, used for reading field values (in the FieldComparators)
     */
    public DocumentComparator(Sort sort, int numberOfHits, IndexReader reader) {
      fieldComparators = new FieldComparator[sort.getSort().length];
      descending = new boolean[sort.getSort().length];
      initializeFieldComparators(sort, numberOfHits, reader, null);
    }

    /**
     * Constructs a <code>DocumentComparator</code> by initializing the
     * {@link org.apache.lucene.search.FieldComparator}s and determining the sort orders.
     *
     * @param sort         The sort used for the creation of the FieldComparators.
     * @param numberOfHits The number of results in the pre-field-collapsed resultset
     * @param reader       The index reader, used for reading field values (in the FieldComparators)
     * @param scores       The scores used for comparing the documents
     */
    public DocumentComparator(Sort sort, int numberOfHits, IndexReader reader, float[] scores) {
      fieldComparators = new FieldComparator[sort.getSort().length];
      descending = new boolean[sort.getSort().length];
      initializeFieldComparators(sort, numberOfHits, reader, scores);
    }

    /**
     * Compares doc1 and doc2 with each other.
     * Compares the two documents with the initialized <code>FieldComparators</code>,
     * if all <code>FieldComparators</code> compares the documents as equal then the document with lowest lucene
     * identifier will be classified as most relevant.
     *
     * @param doc1 The lucene identifier of the first document
     * @param doc2 The lucene identifier of the second document
     * @return -1 if doc1 is less relevant or equal relevant but has a higher lucene id then doc2,
     *         0 if both documents are identical
     *         1 if doc1 is more relevant or equal relevant but has a lower lucene id then doc2
     */
    public int compare(int doc1, int doc2) {
      int result;
      for (int i = 0; i < fieldComparators.length; i++) {
        FieldComparator fieldComparator = fieldComparators[i];
        try {
          scorer.advance(doc1);
          fieldComparator.copy(doc1, doc1);
          scorer.advance(doc2);
          fieldComparator.copy(doc2, doc2);
        } catch (IOException e) {
          throw new RuntimeException(e);
        }
        result = fieldComparator.compare(doc1, doc2);
        result = descending[i] ? result : -result;

        if (result != 0) {
          return result;
        }
      }

      // field comparators identified the field(s) values as equal
      // Document with lowest identifier has higher precendence
      if (doc1 < doc2) {
        return 1;
      } else if (doc1 > doc2) {
        return -1;
      }

      // can only happen if comparing two the exact same document (with same lucene id)
      return 0;
    }

    private void initializeFieldComparators(Sort sort, int numberOfHits, IndexReader indexReader, float[] scores) {
      try {
        scorer = new PredefinedScorer(scores == null ? new float[0] : scores);
        if (sort.getSort().length == 1 && scores != null) {
          fieldComparators[0] = new FloatValueFieldComparator(scores);
          descending[0] = sort.getSort()[0].getReverse();
          return;
        }

        for (int i = 0; i < sort.getSort().length; i++) {
          SortField sortField = sort.getSort()[i];
          fieldComparators[i] = sortField.getComparator(numberOfHits, i);
          descending[i] = sortField.getReverse();
          fieldComparators[i].setNextReader(indexReader, 0);
          fieldComparators[i].setScorer(scorer);
        }
      } catch (IOException e) {
        throw new RuntimeException(e);
      }
    }

  }

  /**
   * Represents a collapse group, that collects statistics for a certain fieldvalue group.
   * <p/>
   * Keeps track of the following statistics during the collapsing on fieldvalue:
   * <ul>
   * <li>how many documents were collapsed
   * <li>what the most relevant head documents are in this group (these will not get collapsed)
   * <li>how many have been procesed
   * </ul>
   */
  private static class CollapsedDocumentGroup {

    int collapsedDocuments;
    int totalCount;
    final DocumentPriorityQueue priorityQueue;

    /**
     * Constructs a <code>CollapsedDocumentGroup</code>. Keeps track of the of the most relevant documents
     * in this group. These documents will stay in the resultset and do not get collaped, no more then the
     * specified collpasThreshold will be kept inside this <code>CollapsedDocumentGroup</code>.
     *
     * @param totalCount         the total amount documents processed in the collapsing process
     * @param collapsedDocuments the amount of documents collapsed under this group
     * @param comparator         The document comparater used inside priority queue
     * @param collapsThreshold   The threshold to start collapsing from
     */
    private CollapsedDocumentGroup(int totalCount, int collapsedDocuments, DocumentComparator comparator, int collapsThreshold) {
      this.totalCount = totalCount;
      this.collapsedDocuments = collapsedDocuments;
      this.priorityQueue = new DocumentPriorityQueue(comparator, collapsThreshold);
    }
  }

  /**
   * A scorer that returns scores from a predefined array of scores.
   */
  private static final class PredefinedScorer extends Scorer {

    private final float[] scores;
    private int index;

    private PredefinedScorer(float[] scores) {
      super(null);
      this.scores = scores;
    }

    /**
     * {@inheritDoc}
     */
    public float score() throws IOException {
      return scores[index];
    }

    /**
     * {@inheritDoc}
     */
    public int docID() {
      return index;
    }

    /**
     * {@inheritDoc}
     */
    public int nextDoc() throws IOException {
      return index < scores.length ? ++index : NO_MORE_DOCS;
    }

    /**
     * {@inheritDoc}
     */
    public int advance(int target) throws IOException {
      if (target < scores.length) {
        index = target;
        return index;
      }
      return NO_MORE_DOCS;
    }

  }

  /**
   * Field comparator for floats. This implementation is a little bit faster (around 30 ms) then {@link FloatComparator.FloatComparator}
   * because it does not copy floats in {@link #copy(int, int)}. This is not necessary in fieldcollapsing case, because
   * the arguments passed to the {@link #compare(int, int)} method are document ids and the values array is sorted on
   * document order.
   */
  private static class FloatValueFieldComparator extends FieldComparator {

    private final float[] values;

    private FloatValueFieldComparator(float[] values) {
      this.values = values;
    }

    /**
     * Compares value for doc1 with value for doc2.
     *
     * @param doc1 first document to compare
     * @param doc2 second document to compare
     * @return any N < 0 if doc2's value is sorted after
     *         doc1, any N > 0 if the doc2's value is sorted before
     *         doc1 and 0 if they are equal
     */
    public int compare(int doc1, int doc2) {
      final float value1 = values[doc1];
      final float value2 = values[doc2];

      if (value1 > value2) {
        return 1;
      } else if (value1 < value2) {
        return -1;
      } else {
        return 0;
      }
    }

    /**
     * Unsupported, is not necessary for field collapsing
     *
     * @param doc The document identifier
     * @throws UnsupportedOperationException
     */
    public void setBottom(int doc) {
      throw new UnsupportedOperationException("Method not implemented");
    }

    /**
     * Unsupported, is not necessary for field collapsing
     *
     * @param doc The document identifier
     * @return nothing
     * @throws UnsupportedOperationException
     */
    public int compareBottom(int doc) throws IOException {
      throw new UnsupportedOperationException("Method not implemented");
    }

    /**
     * {@inheritDoc}
     * <br /><br />
     * Does nothing, necessary to implement because of superclass.
     */
    public void copy(int slot, int doc) throws IOException {
    }

    /**
     * {@inheritDoc}
     * <br /><br />
     * Does nothing, necessary to implement because of superclass.
     */
    public void setNextReader(IndexReader reader, int docBase) throws IOException {
    }

    /**
     * {@inheritDoc}
     */
    public int sortType() {
      return SortField.FLOAT;
    }

    /**
     * Unsupported, is not necessary for field collapsing
     *
     * @param doc The document identifier
     * @throws UnsupportedOperationException
     */
    public Comparable value(int doc) {
      throw new UnsupportedOperationException("Method not implemented");
    }

  }
}
