/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.solr.search.fieldcollapse.collector;

import org.apache.lucene.search.FieldCache;
import org.apache.solr.common.SolrException;
import org.apache.solr.common.params.CollapseParams;
import org.apache.solr.common.util.NamedList;
import org.apache.solr.request.SolrQueryRequest;
import org.apache.solr.schema.FieldType;
import org.apache.solr.search.DocIterator;
import org.apache.solr.search.DocList;
import org.apache.solr.search.SolrIndexSearcher;
import org.apache.solr.search.fieldcollapse.CollapseGroup;
import org.apache.solr.search.fieldcollapse.collector.aggregate.*;
import org.apache.solr.util.plugin.NamedListInitializedPlugin;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A collapse collector factory that creates collapse collectors that create aggregate statistics based on the collapsed
 * documents.
 *
 * <p>
 * If the <code>collapse.aggregate</code> is set with a string that describes what data to collect then a new collapse
 * collector is created that collects this data.
 *
 * <p>
 * The string should be formatted acording to the following format:
 * [function name]([field name])[, ] ...
 * Example: avg(price), max(stock), sum(stock)
 *
 * <p>
 * When this collapse collector factory is configured this feature is <b>not</b> enabled by default.
 */
public class AggregateCollapseCollectorFactory implements CollapseCollectorFactory, NamedListInitializedPlugin {

  private final static Pattern splitPattern = Pattern.compile(",");
  private final static Pattern aggregatePattern = Pattern.compile("(.+)\\((.+)\\)");

  private final Map<String, Class<? extends AggregateFunction>> aggregateFunctions = new HashMap<String, Class<? extends AggregateFunction>>();

  @SuppressWarnings("unchecked")
  public void init(NamedList args) {
    NamedList<String> functionClassNames = (NamedList<String>) args.get("aggregateFunctions");
    if (functionClassNames == null) {
      aggregateFunctions.put("avg", AverageFunction.class);
      aggregateFunctions.put("sum", SumFunction.class);
      aggregateFunctions.put("min", MinFunction.class);
      aggregateFunctions.put("max", MaxFunction.class);
      return;
    }

    for (Map.Entry<String, String> entry : functionClassNames) {
      Class<? extends AggregateFunction> functionClass;
      try {
        functionClass = (Class<? extends AggregateFunction>) Class.forName(entry.getValue());
      } catch (ClassNotFoundException e) {
        throw new RuntimeException(e);
      }
      aggregateFunctions.put(entry.getKey(), functionClass);
    }
  }

  /**
   * Returns an new collapse collector instance that creates aggregate statictics if the collapse.aggregate paramter
   * is set.
   *
   * @param request The specified request
   * @return an new collapse collector instance if the collapse.aggregate parameter is set properly
   */
  public CollapseCollector createCollapseCollector(SolrQueryRequest request) {
    String aggregateParameterValue = request.getParams().get(CollapseParams.COLLAPSE_AGGREGATE);
    if (aggregateParameterValue == null) {
      return null;
    }

    Map<AggregateField, AggregateFunction> aggregateFunctions = resolveAggregateFunctions(aggregateParameterValue);
    return new AggregateCollapseCollector(aggregateFunctions, request.getSearcher());
  }

  /**
   * Resolves on a per field basis the aggregate function.
   * <ul>
   * <li> Splits the functions. E.g. collapse.aggregate=avg(price),sum(stock) into avg(price) and sum(stock)
   * <li> Splits the functionname and fieldname from each function in order to resolve and instantiates the aggregate function.
   * E.g. functionname = avg, fieldname = price
   * <li> Looks up the function class and creates an instance of it.
   * </ul>
   *
   * @param aggregateParameterValue The <code>collapse.aggregate</code> parameter value in the solr request
   * @return an aggragate function per field stored in a map.
   */
  protected Map<AggregateField, AggregateFunction> resolveAggregateFunctions(String aggregateParameterValue) {
    Map<AggregateField, AggregateFunction> fieldNameAggregateFunction = new HashMap<AggregateField, AggregateFunction>();

    String[] aggregates = splitPattern.split(aggregateParameterValue);
    for (String aggregate : aggregates) {
      Matcher matcher = aggregatePattern.matcher(aggregate);
      if (!matcher.matches() || matcher.groupCount() != 2) {
        throw new SolrException(SolrException.ErrorCode.BAD_REQUEST, "Aggregate function syntax is incorrect.");
      }

      String functionName = matcher.group(1).trim();
      String fieldName = matcher.group(2);
      AggregateFunction function = resolveAggregateFunction(functionName);
      fieldNameAggregateFunction.put(new AggregateField(fieldName, functionName), function);
    }

    return fieldNameAggregateFunction;
  }

  /**
   * Returns a new instance of the <code>AggregateFunction</code> with the specified functionName.
   *
   * @param functionName The name of the function to create a new instance of
   * @return a new instance of the <code>AggregateFunction</code> with the specified functionName
   * @throws UnsupportedOperationException When no aggregate functions with the specified function name exists
   */
  protected AggregateFunction resolveAggregateFunction(String functionName) {
    Class<? extends AggregateFunction> classOfFunction = aggregateFunctions.get(functionName);
    if (classOfFunction == null) {
      throw new SolrException(SolrException.ErrorCode.BAD_REQUEST, String.format("Aggregate function with name [%s] does not exist", functionName));
    }

    try {
      return classOfFunction.newInstance();
    } catch (InstantiationException e) {
      throw new RuntimeException(e);
    } catch (IllegalAccessException e) {
      throw new RuntimeException(e);
    }
  }


  //============================================== Inner classes =======================================================

  private static class AggregateCollapseCollector extends AbstractCollapseCollector {

    private final Map<AggregateField, AggregateFunction> functions;
    private final Map<String, FieldCache.StringIndex> fieldCaches = new HashMap<String, FieldCache.StringIndex>();
    private final Map<String, FieldType> fieldTypes = new HashMap<String, FieldType>();

    private AggregateCollapseCollector(Map<AggregateField, AggregateFunction> functions, SolrIndexSearcher searcher) {
      super(searcher);
      this.functions = functions;

      for (AggregateField aggregateField : functions.keySet()) {
        try {
          String fieldName = aggregateField.getFieldName();
          fieldCaches.put(fieldName, FieldCache.DEFAULT.getStringIndex(searcher.getReader(), fieldName));
          fieldTypes.put(fieldName, searcher.getSchema().getFieldType(fieldName));
        } catch (IOException e) {
          throw new RuntimeException(e);
        }
      }
    }

    public void documentCollapsed(int docId, CollapseGroup collapseGroup, CollapseContext collapseContext) {
      for (AggregateField aggregateField : functions.keySet()) {
        FieldCache.StringIndex stringIndex = fieldCaches.get(aggregateField.getFieldName());
        AggregateFunction function = functions.get(aggregateField);

        String fieldCacheValue = stringIndex.lookup[stringIndex.order[docId]];
        String value = fieldTypes.get(aggregateField.getFieldName()).indexedToReadable(fieldCacheValue);
        function.collect(collapseGroup, value);
      }
    }

    public void documentHead(int docHeadId, CollapseGroup collapseGroup, CollapseContext collapseContext) {
    }

    public void getResult(NamedList result, DocList docs, CollapseContext collapseContext) {
      Map<Integer, CollapseGroup> docHeadCollapseGroupAssoc = collapseContext.getDocumentHeadCollapseGroupAssociation();

      for (DocIterator i = docs.iterator(); i.hasNext();) {
          int id = i.nextDoc();
          CollapseGroup collapseGroup = docHeadCollapseGroupAssoc.get(id);
          if (collapseGroup == null) {
            continue;
          }

          for (AggregateField aggregateField : functions.keySet()) {
            AggregateFunction function = functions.get(aggregateField);
            String functionResult = function.calculate(collapseGroup);
            if (functionResult != null) {
              NamedList collapsedGroupResult = getCollapseGroupResult(id, result);
              NamedList aggragateFunctions = (NamedList) collapsedGroupResult.get("aggregate");
              if (aggragateFunctions == null) {
                aggragateFunctions = new NamedList<Object>();
                collapsedGroupResult.add("aggregate", aggragateFunctions);
              }

              aggragateFunctions.add(aggregateField.getUniqueName(), functionResult);
            }
          }

      }
    }

  }

  private final static class AggregateField {

    private final String fieldName;
    private final String functionName;

    private AggregateField(String fieldName, String functionName) {
      this.fieldName = fieldName;
      this.functionName = functionName;
    }

    public String getFieldName() {
      return fieldName;
    }

    public String getFunctionName() {
      return functionName;
    }

    public String getUniqueName() {
      return String.format("%s(%s)", functionName, fieldName);
    }

    @Override
    public boolean equals(Object o) {
      if (this == o) return true;
      if (o == null || getClass() != o.getClass()) return false;

      AggregateField that = (AggregateField) o;

      if (fieldName != null ? !fieldName.equals(that.fieldName) : that.fieldName != null) return false;
      if (functionName != null ? !functionName.equals(that.functionName) : that.functionName != null) return false;

      return true;
    }

    @Override
    public int hashCode() {
      int result = fieldName != null ? fieldName.hashCode() : 0;
      result = 31 * result + (functionName != null ? functionName.hashCode() : 0);
      return result;
    }
  }

}
