/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.solr.search.fieldcollapse.collector;

import org.apache.solr.search.fieldcollapse.CollapseGroup;

import java.util.HashMap;
import java.util.Map;

/**
 * A collapse context allows <code>CollapseCollectors</code> to share data between each other.
 * Per search request only one <code>CollapseContext</code> will be created.
 * A <code>CollapseContext</code> is <b>not</b> thread save.
 */
public class CollapseContext {

  private static final String DOCUMENT_HEAD_COLLAPSE_GROUP_ASSOCIATION_KEY = "documentHeadCollapseGroupAssociation";

  private final Map<String, Object> data = new HashMap<String, Object>();

  public CollapseContext() {
    data.put(DOCUMENT_HEAD_COLLAPSE_GROUP_ASSOCIATION_KEY, new HashMap<Integer, CollapseGroup>());
  }

  /**
   * Sets a value in the collapse context under the specified key.
   *
   * @param key The key
   * @param value The value to share
   */
  public void put(String key, Object value) {
    data.put(key, value);
  }

  /**
   * Returns a value in the collapse context that is associated with the specified key or <code>null</code> if
   * no value is associated with the specified key.
   *
   * @param key The key
   * @return a value in the collapse or <code>null</code>
   */
  public Object get(String key) {
    return data.get(key);
  }

  /**
   * Returns the documenthead collapse group association.
   * The documenthead collapse group association is a map with the document head as key and the collapse group as value.
   * This method never returns a <code>null</code> value.
   *
   * @return the documenthead collapse group association
   */
  public Map<Integer, CollapseGroup> getDocumentHeadCollapseGroupAssociation() {
    return (Map<Integer, CollapseGroup>) data.get(DOCUMENT_HEAD_COLLAPSE_GROUP_ASSOCIATION_KEY);
  }

}
