/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.solr.search.fieldcollapse.collector;

import org.apache.lucene.document.Document;
import org.apache.lucene.document.Fieldable;
import org.apache.lucene.search.DocIdSetIterator;
import org.apache.lucene.util.OpenBitSet;
import org.apache.solr.common.SolrDocument;
import org.apache.solr.common.SolrDocumentList;
import org.apache.solr.common.params.CollapseParams;
import org.apache.solr.common.params.SolrParams;
import org.apache.solr.common.util.NamedList;
import org.apache.solr.request.SolrQueryRequest;
import org.apache.solr.schema.IndexSchema;
import org.apache.solr.search.DocIterator;
import org.apache.solr.search.DocList;
import org.apache.solr.search.SolrIndexSearcher;
import org.apache.solr.search.fieldcollapse.CollapseGroup;
import org.apache.solr.update.DocumentBuilder;

import java.io.IOException;
import java.util.*;
import java.util.regex.Pattern;

/**
 * A collector factory that creates collapse collectors that collect predefined fieldvalues from collapsed documents.
 * In order to enable this specify collapse.includeCollapsedDocs.fl=[val] in the request. Where [val] may be a star (*)
 * to include all field of the collapsed documents or a comma separated list of the collapsed documents' fieldnames to include.
 * When configured this feature is <b>not</b> enabled by default. 
 */
public class DocumentFieldsCollapseCollectorFactory implements CollapseCollectorFactory {

  private static final Pattern splitPattern = Pattern.compile(",");

  /**
   * {@inheritDoc}
   */
  public CollapseCollector createCollapseCollector(SolrQueryRequest request) {
    SolrParams params = request.getParams();
    String collapsedDocumentFields = params.get(CollapseParams.COLLAPSE_INCLUDE_COLLAPSED_DOCS_FIELDS);
    if (collapsedDocumentFields == null || "".equals(collapsedDocumentFields.trim())) {
      return null;
    }

    Set<String> fields;
    if ("*".equals(collapsedDocumentFields.trim())) {
      fields = null;
    } else {
      fields = new HashSet<String>(Arrays.asList(splitPattern.split(collapsedDocumentFields)));
    }

    return new CollapsedDocumentCollapseCollector(request.getSearcher(), fields);
  }


  //============================================== Inner classes =======================================================

  private static class CollapsedDocumentCollapseCollector extends AbstractCollapseCollector {

    private final Map<CollapseGroup, OpenBitSet> collapsedDocumentsByFieldValue = new HashMap<CollapseGroup, OpenBitSet>();
    private final Set<String> includeCollapsedDocumentsFields;

    public CollapsedDocumentCollapseCollector(SolrIndexSearcher searcher, Set<String> fields) {
      super(searcher);
      this.includeCollapsedDocumentsFields = fields;
    }

    public void documentCollapsed(int docId, CollapseGroup collapseGroup, CollapseContext collapseContext) {
      OpenBitSet bitSet = collapsedDocumentsByFieldValue.get(collapseGroup);
      if (bitSet == null) {
        collapsedDocumentsByFieldValue.put(collapseGroup, bitSet = new OpenBitSet());
      }
      // We cannot use fastset because we do not know the size upfront.
      // The set method is more expensive than fastset, but that is the best for now.
      // Setting bitSet to maxDoc may increase the memory footprint dramatically, because for each distinct fieldvalue
      // there is a bitset
      bitSet.set(docId);
    }

    public void documentHead(int docHeadId, CollapseGroup collapseGroup, CollapseContext collapseContext) {
    }

    @SuppressWarnings("unchecked")
    public void getResult(NamedList result, DocList docs, CollapseContext collapseContext) {
      IndexSchema schema = searcher.getSchema();
      DocumentBuilder documentBuilder = new DocumentBuilder(schema);
      Map<Integer, CollapseGroup> docHeadCollapseGroupAssoc = collapseContext.getDocumentHeadCollapseGroupAssociation();

      try {
        for (DocIterator i = docs.iterator(); i.hasNext();) {
          int id = i.nextDoc();
          if (!docHeadCollapseGroupAssoc.containsKey(id)) {
            continue;
          }

          CollapseGroup collapseGroup = docHeadCollapseGroupAssoc.get(id);
          OpenBitSet bitSet = collapsedDocumentsByFieldValue.get(collapseGroup);
          SolrDocumentList documentList = new SolrDocumentList();
          for (DocIdSetIterator iterator = bitSet.iterator(); iterator.nextDoc() != DocIdSetIterator.NO_MORE_DOCS;) {
            Document luceneDocument = searcher.doc(iterator.docID(), includeCollapsedDocumentsFields);

            // the luceneDocument may still contain fields that are not specified in the includeCollapsedDocumentsFields
            // this is because the document was read from the document cache. Off course we only want to return requested the fields.
            // There is also another doc() method that accepts a FieldSelector instead of a list of fields, but that
            // method does not use the Solr document cache, which from my perspective is a necessity
            if (includeCollapsedDocumentsFields != null && luceneDocument.getFields().size() > includeCollapsedDocumentsFields.size()) {
              Document document = new Document();
              for (Object object : luceneDocument.getFields()) {
                Fieldable field = (Fieldable) object;
                if (includeCollapsedDocumentsFields.contains(field.name())) {
                  document.add(field);
                }
              }
              luceneDocument = document;
            }

            SolrDocument solrDocument = new SolrDocument();
            documentBuilder.loadStoredFields(solrDocument, luceneDocument);
            documentList.add(solrDocument);
          }
          documentList.setNumFound(bitSet.cardinality());
          NamedList collapsedGroupResult = getCollapseGroupResult(id, result);
          collapsedGroupResult.add("collapsedDocs", documentList);
        }
      } catch (IOException e) {
        throw new RuntimeException(e);
      }
    }
  }

}
