/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.solr.search.fieldcollapse.collector;

import org.apache.solr.common.params.CollapseParams;
import org.apache.solr.common.params.SolrParams;
import org.apache.solr.common.util.NamedList;
import org.apache.solr.request.SolrQueryRequest;
import org.apache.solr.search.DocIterator;
import org.apache.solr.search.DocList;
import org.apache.solr.search.SolrIndexSearcher;
import org.apache.solr.search.fieldcollapse.CollapseGroup;
import org.apache.solr.search.fieldcollapse.util.Counter;

import java.util.HashMap;
import java.util.Map;

/**
 * A collapse collector factory that creates collapse collectors that the collapse counts per document group and
 * return the counts in the response per collapsed group most relevant document id.
 * When this collapse collector factory is configured then this feature is enabled by default.
 * To disable this for a search add collapse.info.doc=false in the request.
 */
public class DocumentGroupCountCollapseCollectorFactory implements CollapseCollectorFactory {

  /**
   * {@inheritDoc}
   */
  public CollapseCollector createCollapseCollector(SolrQueryRequest request) {
    SolrParams params = request.getParams();
    boolean includeDocCount = params.getBool(CollapseParams.COLLAPSE_INFO_DOC, true);
    return includeDocCount ? new DocumentCountCollapseCollector(request.getSearcher()) : null;
  }


  //============================================== Inner classes =======================================================

  private static class DocumentCountCollapseCollector extends AbstractCollapseCollector {

    private final Map<CollapseGroup, Counter> fieldValueCount = new HashMap<CollapseGroup, Counter>();
    private final Map<Integer, Counter> documentHeadCount = new HashMap<Integer, Counter>();

    public DocumentCountCollapseCollector(SolrIndexSearcher searcher) {
      super(searcher);
    }

    public void documentCollapsed(int docId, CollapseGroup collapseGroup, CollapseContext collapseContext) {
      Counter counter = fieldValueCount.get(collapseGroup);
      if (counter == null) {
        counter = new Counter(0);
        fieldValueCount.put(collapseGroup, counter);
      }
      counter.increment();
    }

    public void documentHead(int docHeadId, CollapseGroup collapseGroup, CollapseContext collapseContext) {
      documentHeadCount.put(docHeadId, fieldValueCount.get(collapseGroup));
      collapseContext.put("documentHeadCount", documentHeadCount);
    }

    public void getResult(NamedList result, DocList docs, CollapseContext collapseContext) {
      for (DocIterator i = docs.iterator(); i.hasNext();) {
        int id = i.nextDoc();
        Counter counter = documentHeadCount.get(id);
        if (counter == null) {
          continue;
        }

        NamedList collapsedGroupResult = getCollapseGroupResult(id, result);
        collapsedGroupResult.add("collapseCount", counter.getCount());
      }
    }
  }

}
