/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.solr.search.fieldcollapse.collector;

import org.apache.solr.common.params.CollapseParams;
import org.apache.solr.common.params.SolrParams;
import org.apache.solr.common.util.NamedList;
import org.apache.solr.request.SolrQueryRequest;
import org.apache.solr.schema.FieldType;
import org.apache.solr.schema.IndexSchema;
import org.apache.solr.search.DocIterator;
import org.apache.solr.search.DocList;
import org.apache.solr.search.SolrIndexSearcher;
import org.apache.solr.search.fieldcollapse.CollapseGroup;
import org.apache.solr.search.fieldcollapse.util.Counter;

import java.util.HashMap;
import java.util.Map;

/**
 * A collapse collector factory that creates collapse collectors that collect t the collapse count per collapsed group
 * and return the counts in the response per collepsed group field value. When this collapse collector factory is
 * configured it is enabled by default. To disable this feature for a certain search add collapse.info.count=false
 * to the request as parameter.
 */
public class FieldValueCountCollapseCollectorFactory implements CollapseCollectorFactory {

  /**
   * {@inheritDoc}
   */
  public CollapseCollector createCollapseCollector(SolrQueryRequest request) {
    SolrParams params = request.getParams();
    String collapseField = params.required().get(CollapseParams.COLLAPSE_FIELD);
    boolean includeFieldValueCount = params.getBool(CollapseParams.COLLAPSE_INFO_COUNT, true);
    return includeFieldValueCount ? new FieldValueCountCollapseCollector(request.getSearcher(), collapseField) : null;
  }


  //============================================== Inner classes =======================================================

  private static class FieldValueCountCollapseCollector extends AbstractCollapseCollector {

    private final FieldType collapseFieldType;
    private final Map<CollapseGroup, Counter> fieldValueCount = new HashMap<CollapseGroup, Counter>();

    public FieldValueCountCollapseCollector(SolrIndexSearcher searcher, String collapseField) {
      super(searcher);
      IndexSchema schema = searcher.getSchema();
      collapseFieldType = schema.getField(collapseField).getType();
    }

    public void documentCollapsed(int docId, CollapseGroup collapseGroup, CollapseContext collapseContext) {
      Counter counter = fieldValueCount.get(collapseGroup);
      if (counter == null) {
        counter = new Counter(0);
        fieldValueCount.put(collapseGroup, counter);
      }
      counter.increment();
    }

    public void documentHead(int docHeadId, CollapseGroup collapseGroup, CollapseContext collapseContext) {
    }

    public void getResult(NamedList result, DocList docs, CollapseContext collapseContext) {
      Map<Integer, CollapseGroup> docHeadCollapseGroupAssoc = collapseContext.getDocumentHeadCollapseGroupAssociation();
      for (DocIterator i = docs.iterator(); i.hasNext();) {
        int id = i.nextDoc();
        CollapseGroup collapseGroup = docHeadCollapseGroupAssoc.get(id);
        if (collapseGroup == null) {
          continue;
        }

        NamedList collapsedGroupResult = getCollapseGroupResult(id, result);
        String fieldValue = collapseFieldType.indexedToReadable(collapseGroup.getKey());
        collapsedGroupResult.add("fieldValue", fieldValue);
      }
    }

  }
}
