/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.solr.util;

import org.apache.lucene.index.IndexReader;
import org.apache.lucene.search.Scorer;
import org.apache.lucene.search.Filter;
import org.apache.lucene.util.OpenBitSet;
import org.apache.solr.search.BitDocSet;
import org.apache.solr.search.DocSet;
import org.apache.solr.search.DocSetAwareCollector;
import org.apache.solr.search.DocIterator;

import java.io.IOException;

/**
 * Collects the documents with the score.
 */
public class DocSetScoreCollector extends DocSetAwareCollector {
  private final float[] scores;
  private final OpenBitSet bits;

  private Scorer scorer = null;
  private int docBase = 0;

  /**
   * Constructs a DocSetHitCollector with the specified parameter.
   *
   * @param maxDoc The maximum size of all the documents in the index
   */
  public DocSetScoreCollector(int maxDoc) {
    scores = new float[maxDoc];
    bits = new OpenBitSet(maxDoc);
  }

  /**
   * Collects documents as specified in {@link super#collect(int)}.
   * Also stores the score associated with this document.
   *
   * @param doc The document id to collect
   * @throws IOException
   */
  public void collect(int doc) throws IOException {
    doc = doc + docBase;
    bits.fastSet(doc);
    scores[doc] = scorer.score();
  }

 /**
  * {@inheritDoc}
  */
  public boolean acceptsDocsOutOfOrder() {
    return false;
  }

  // ================================================= Setter/Getter ===================================================

  /**
   * {@inheritDoc}
   */
  public void setNextReader(IndexReader reader, int docBase) throws IOException {
    this.docBase = docBase;
  }

  /**
   * {@inheritDoc}
   */
  public void setScorer(Scorer scorer) throws IOException {
    this.scorer = scorer;
  }

  /**
   * Returns a docset of the collected documents.
   *
   * @return a docset of the collected documents
   */
  public DocSet getDocSet() {
    return new DelegateDocSet(new BitDocSet(bits), scores);
  }

  /**
   * Returns the scores for the collected documents.
   * The index of the score represents the lucene document identifier.
   *
   * @return the scores for the collected documents
   */
  public float[] getScores() {
    return scores;
  }


  // ============================================ Inner Classes ======================================================

  /**
   * A delegate docset that also has the associated score for each document in the contained docset.
   */
  public class DelegateDocSet implements DocSet {

    private final DocSet docSet;
    private final float[] scores;

    public DelegateDocSet(DocSet docSet, float[] scores) {
      this.docSet = docSet;
      this.scores = scores;
    }

    public void add(int doc) {
      docSet.add(doc);
    }

    public void addUnique(int doc) {
      docSet.addUnique(doc);
    }

    public int size() {
      return docSet.size();
    }

    public boolean exists(int docid) {
      return docSet.exists(docid);
    }

    public DocIterator iterator() {
      return docSet.iterator();
    }

    public OpenBitSet getBits() {
      return docSet.getBits();
    }

    public long memSize() {
      return docSet.memSize();
    }

    public DocSet intersection(DocSet other) {
      return docSet.intersection(other);
    }

    public int intersectionSize(DocSet other) {
      return docSet.intersectionSize(other);
    }

    public DocSet union(DocSet other) {
      return docSet.union(other);
    }

    public int unionSize(DocSet other) {
      return docSet.unionSize(other);
    }

    public DocSet andNot(DocSet other) {
      return docSet.andNot(other);
    }

    public int andNotSize(DocSet other) {
      return docSet.andNotSize(other);
    }

    public Filter getTopFilter() {
      return docSet.getTopFilter();
    }

    public float[] getScores() {
      return scores;
    }
  }

}
